"""Httpr module for handling http requests and responses."""

import re
import warnings
from datetime import datetime
from typing import Any, Dict, List, Optional, Union
from urllib.parse import urljoin, urlparse, urlunparse  # noqa:F401

import requests
from bs4 import BeautifulSoup
from requests import Response
from requests.packages.urllib3.exceptions import InsecureRequestWarning


# Suppress insecure request warnings
warnings.simplefilter("ignore", InsecureRequestWarning)


# Constants
NOW = datetime.now()
USER_AGENT_LIST = [
    (
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko)"
        " Chrome/48.0.2564.116 Safari/537.36"
    ),
    (
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_5) AppleWebKit/605.1.15 (KHTML,"
        " like Gecko) Version/13.1.1 Safari/605.1.15"
    ),
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:77.0) Gecko/%s Firefox/77.0"
    % NOW.strftime("%Y%m%d"),
    (
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_5) AppleWebKit/537.36 (KHTML,"
        " like Gecko) Chrome/83.0.4103.97 Safari/537.36"
    ),
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:77.0) Gecko/%s Firefox/77.0"
    % NOW.strftime("%Y%m%d"),
    (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like"
        " Gecko) Chrome/83.0.4103.97 Safari/537.36"
    ),
]
DEFAULT_USER_AGENT = USER_AGENT_LIST[0]
DEFAULT_REQUEST_HEADERS = {
    "Accept": "*/*",
    "Accept-Language": "en-US,en;q=0.8",
    "Cache-Control": "max-age=0",
    "Connection": "keep-alive",
    "User-Agent": DEFAULT_USER_AGENT,
}
META_REDIRECT_REGEX = re.compile(
    "<meta[^>]*?url=(.*?)[\"']", re.IGNORECASE
)  # noqa: F841


def filter_headers(
    headers: Dict[str, str], keys: Optional[List[str]] = None
) -> Dict[str, str]:
    """Filter headers by key.

    Args:
        headers (Dict[str, str]): headers to filter
        keys (Optional[List[str]]): keys to filter by. Defaults to None.

    Returns:
        Dict[str, str]: filtered headers
    """
    if keys is None:
        keys = []

    filtered = {}
    if len(keys) > 0:
        for header, value in headers.items():
            for response_header in keys:
                if header.lower() == response_header.lower():
                    filtered[header] = value
    else:
        filtered = headers

    return filtered


def get_http_version(raw_version: int) -> str:
    """Given the raw.version, return the HTTP version as decimal string.

    Args:
        raw_version (int): raw version

    Returns:
        str: 1.0, 1.1, 1.2, 1.3, 2.0, etc.
    """
    splits = list(str(raw_version))
    splits.insert(1, ".")
    version = "".join(splits)
    return version


def get_http_version_text(raw_version: int) -> str:
    """Given the raw.version, return the HTTP version as decimal string.

    Args:
        raw_version (int): raw version

    Returns:
        str: HTTP/1.0, HTTP/1.1, HTTP/1.2, HTTP/1.3, etc.
    """
    tls_version = get_http_version(raw_version)
    return "HTTP/%s" % tls_version


def build_full_url(url: str, next_url: str) -> Optional[str]:
    """Builds a full URL from a relative URL.

    Args:
        url (str): url to build from
        next_url (str): relative url to build

    Returns:
        Optional[str]: full url
    """
    if next_url == "":
        return None

    uparsed = urlparse(url)
    if next_url.startswith("http"):
        return next_url
    elif next_url.startswith("//"):
        return uparsed.scheme + ":" + next_url
    elif next_url.startswith("/"):
        return uparsed.scheme + "://" + uparsed.netloc + next_url
    else:
        # TODO: Doesn't handle relative pathnames correctly
        return urljoin(url, next_url)


def extract_redirects_from_html_meta(soup: BeautifulSoup) -> Optional[str]:
    """Extracts HTML meta redirects from a BeautifulSoup object.

    Args:
        soup (BeautifulSoup): BeautifulSoup object

    Returns:
        Optional[str]: redirect url
    """
    meta_redirect = soup.find("meta", attrs={"http-equiv": "refresh"})
    if meta_redirect is not None:
        content = meta_redirect.get("content", "")
        if content != "":
            t = re.split(re.compile(r"url=", re.I), content)[1]
            t = re.sub(r'["\']', "", t).strip()
            return t

    return None


def make_request(
    method: str,
    url: str,
    **kwargs: Dict[str, Any],
) -> List[Union[str, Union[Response, Exception]]]:
    """Make a request to a given URL."""
    # The result will be a tuple of the URL and the response object,
    # or the URL and the error that's raised.
    try:
        with requests.request(method, url, **kwargs) as res:
            return [url, res]
    except Exception as e:
        # TODO: Handle this better
        return [url, e]


def get_next_url(res: Union[Response, Exception]) -> Optional[str]:
    """Get the next URL from a response.

    Args:
        res (Union[Response, Exception]): response object

    Returns:
        Optional[str]: next url
    """
    if isinstance(res, Exception):
        return None
    else:
        for key in res.headers.keys():
            if key.lower() == "location":
                return res.headers[key]

        else:
            return None


def build_redirect_chain(
    method: str,
    url: str,
    timeout: Optional[int] = 30,
    headers: Optional[Dict[str, str]] = None,
    proxies: Optional[Dict[str, str]] = None,
    follow_meta: bool = True,
    **kwargs: Dict[str, Any],
) -> List[Union[str, List[Union[str, Union[Response, Exception]]]]]:
    """Build a chain of HTTP redirects for a given URL."""
    chains = []
    current_url = url  # type: Optional[str]

    while current_url is not None:
        chain = [current_url, None]
        try:
            headers = {
                **DEFAULT_REQUEST_HEADERS,
                **(headers or {}),
            }
            chain = make_request(
                method,
                current_url,
                proxies=proxies,
                timeout=timeout,
                headers=headers,
                allow_redirects=False,
                verify=False,
                **kwargs,
            )
            res = chain[1]  # type: Union[Response, Exception]

            next_url = get_next_url(res)

            if next_url is None or next_url == "":
                if follow_meta is True:  # pragma: no cover
                    if isinstance(res, Exception) is False:
                        content_type = res.headers.get("Content-Type", "")

                        if "html" in content_type or "plain" in content_type:
                            content = res.text
                            soup = BeautifulSoup(content, "html.parser")
                            next_url = extract_redirects_from_html_meta(soup)

                            if next_url is not None:
                                current_url = build_full_url(
                                    current_url, next_url
                                )
                            else:
                                current_url = None
                        else:
                            current_url = None
                    else:
                        current_url = None
                else:
                    current_url = None
            else:
                current_url = build_full_url(current_url, next_url)
        except Exception as e:
            chain[1] = e
            current_url = None

        chains.append(chain)

    return chains



"""Command-line script for checking url aliveness and status."""

import re
import sys
from typing import Tuple

import click
import requests

from .commons import common_options, extract_urls, parse_input_methods
from .constants import (  # URL_REGEX,
    DEFAULT_CATEGORIZED_HEADERS,
    HELP_SHORT_TEXT,
    NO_ARGS_TEXT,
)
from .exceptions import (
    REQUESTS_CONNECTION_ERROR_MESSAGES,
    REQUESTS_SSL_ERROR_MESSAGE,
    REQUESTS_TIMEOUT_ERROR_MESSAGE,
    REQUESTS_TOO_MANY_REDIRECTS_ERROR_MESSAGE,
    REQUESTS_UNHANDLED_CONNECTION_ERROR_MESSAGE,
)
from .httpr import build_redirect_chain, filter_headers, get_http_version_text


# Initialize global variables
HEADER_KEY_TRUNC_LENGTH = 70
QUIET_MODE = False
OUTPUT_FILE = None


@common_options(
    cmd_type=click.command,
    name="urlcheck",
    description="Check url(s) for their aliveness and status.",
    version="0.1.0",
)
@click.option(
    "-t",
    "--no-truncate",
    "no_truncate",
    is_flag=True,
    help="Disable header truncation.",
    default=False,
)
@click.option(
    "-S",
    "--show-headers",
    "show_headers",
    is_flag=True,
    help="Disable header truncation.",
    default=False,
)
@click.pass_context
def cli(  # noqa: C901
    ctx: click.Context,
    values: Tuple[str, ...],
    interactive: bool,
    no_truncate: bool,
    show_headers: bool,
) -> None:
    """Check whois on domains and ip addresses."""
    args = parse_input_methods(values, interactive, ctx)

    if len(args) == 0:
        click.echo("Error: %s" % NO_ARGS_TEXT, err=True)
        click.echo(HELP_SHORT_TEXT.format(name=ctx.command.name), err=True)
        sys.exit(1)

    urls = []
    for arg in args:
        for url in extract_urls(arg):
            if url not in urls:
                urls.append(url)

    for u in range(len(urls)):
        url = urls[u]
        results = build_redirect_chain("GET", url, 30, {}, None, True)

        for r in range(len(results)):
            url, response = results[r]

            padding = 3
            # Print the URL
            if r <= 0:
                click.echo("->", nl=False)
            else:
                padding = 3
                click.echo(">>", nl=False)

            click.echo(" ", nl=False)
            click.echo(url, nl=False)
            click.echo()

            if issubclass(type(response), Exception):
                click.echo(" " * padding, nl=False, err=True)

                if isinstance(response, requests.exceptions.SSLError):
                    click.echo(REQUESTS_SSL_ERROR_MESSAGE, err=True)
                elif isinstance(response, requests.exceptions.Timeout):
                    click.echo(REQUESTS_TIMEOUT_ERROR_MESSAGE, err=True)
                elif isinstance(response, requests.exceptions.TooManyRedirects):
                    click.echo(
                        REQUESTS_TOO_MANY_REDIRECTS_ERROR_MESSAGE, err=True
                    )
                elif isinstance(response, requests.exceptions.ConnectionError):
                    for (
                        search_text,
                        error_message,
                    ) in REQUESTS_CONNECTION_ERROR_MESSAGES.items():
                        if re.search(search_text, str(response)) is not None:
                            click.echo(error_message, err=True)
                            break

                    else:
                        click.echo(
                            REQUESTS_UNHANDLED_CONNECTION_ERROR_MESSAGE,
                            err=True,
                        )
                else:
                    click.echo(str(response), err=True)

                continue
            else:
                # Print the response status
                http_version = get_http_version_text(response.raw.version)
                status_code = response.status_code
                reason = response.reason

                click.echo(" " * padding, nl=False)
                click.echo("%s - %i - " % (http_version, status_code), nl=False)
                if 100 <= status_code < 200:
                    click.echo(reason)
                elif 200 <= status_code < 300:
                    click.echo(reason)
                elif 300 <= status_code < 400:
                    click.echo(reason)
                elif 400 <= status_code < 500:
                    click.echo(reason)
                else:
                    click.echo(reason)

                # Print the response headers
                headers = response.headers
                if show_headers is True:
                    headers = filter_headers(response.headers, [])
                else:
                    headers = filter_headers(
                        response.headers,
                        [
                            v
                            for vv in DEFAULT_CATEGORIZED_HEADERS.values()
                            for v in vv
                        ],
                    )

                for key, val in headers.items():
                    click.echo(" " * padding, nl=False)
                    click.echo(key, nl=False)
                    click.echo(": ", nl=False)
                    if (
                        len(val) > HEADER_KEY_TRUNC_LENGTH
                        and no_truncate is False
                    ):
                        val = val[:HEADER_KEY_TRUNC_LENGTH] + "..."

                    click.echo(val)

        # Add a newline between URLs, but not after the last one
        if u < len(urls) - 1:
            click.echo("")


if __name__ == "__main__":  # pragma: no cover
    cli()



"""Module for IP address manipulation."""

import ipaddress
from typing import Any, Dict, List, Optional

import requests

from .cache import cache
from .constants import DEFAULT_REQUEST_TIMEOUT


# Global constants
PRIVATE_IP_CIDR_RANGES = [
    "10.0.0.0/8",  # Private-Use Networks - RFC1918
    "172.16.0.0/12",  # Private-Use Networks - RFC1918
    "192.168.0.0/16",  # Private-Use Networks - RFC1918
    "100.64.0.0/10",  # Shared Address Space - RFC6598
    "fc00::/7",  # IPv6 unique local addr (ULA) - RFC4193
]
TOR_PROJECT_NODE_ENDPOINT = (
    "https://check.torproject.org/cgi-bin/TorBulkExitList.py"
)
IPINFO_API_ENDPOINT = "https://ipinfo.io/%s/json"
# https://docs.aws.amazon.com/vpc/latest/userguide/aws-ip-ranges.html#aws-ip-download
AWS_IP_RANGES_ENDPOINT = "https://ip-ranges.amazonaws.com/ip-ranges.json"
CLOUDFLARE_IPS_BASE_ENDPOINT = "https://www.cloudflare.com/ips-v%i/#"
CLOUDFLARE_IPV4_RANGES_ENDPOINT = CLOUDFLARE_IPS_BASE_ENDPOINT % 4
CLOUDFLARE_IPV6_RANGES_ENDPOINT = CLOUDFLARE_IPS_BASE_ENDPOINT % 6
FASTLY_IP_RANGES_ENDPOINT = "https://api.fastly.com/public-ip-list"


def is_ipv4_addr(ipaddr: str) -> bool:
    """Check if a given ip addr is a valid ipv4 addr.

    Args:
        ipaddr (str): IP address to check.

    Returns:
        bool: True if ipaddr is a valid ipv4 addr.
    """
    if isinstance(ipaddr, str):
        try:
            ip = ipaddress.ip_address(ipaddr)
            return isinstance(ip, ipaddress.IPv4Address)
        except ValueError:
            return False

    return False


def is_ipv6_addr(ipaddr: str) -> bool:
    """Check if a given ip addr is a valid ipv6 addr.

    Args:
        ipaddr (str): IP address to check.

    Returns:
        bool: True if ipaddr is a valid ipv6 addr.
    """
    if isinstance(ipaddr, str):
        try:
            ip = ipaddress.ip_address(ipaddr)
            return isinstance(ip, ipaddress.IPv6Address)
        except ValueError:
            return False

    return False


def is_valid_ip_addr(ipaddr: str) -> bool:
    """Check if a given ip addr is a valid ipv4 or ipv6 addr.

    Args:
        ipaddr (str): IP address to check.

    Returns:
        bool: True if ipaddr is a valid ipv4 or ipv6 addr.
    """
    return any(
        [
            is_ipv4_addr(ipaddr),
            is_ipv6_addr(ipaddr),
        ]
    )


def is_private_ip(ipaddr: str) -> bool:
    """Check if ip addr is a private addr.

    Args:
        ipaddr (str): IP address to check.

    Returns:
        bool: True if ipaddr is a private addr.
    """
    try:
        ip = ipaddress.ip_address(ipaddr)
        return any(
            ip in ipaddress.ip_network(cidr) for cidr in PRIVATE_IP_CIDR_RANGES
        )
    except ValueError:
        return False


def get_net_size(cidr: str) -> int:
    """Get network size from cidr notation.

    Args:
        cidr (str): CIDR notation.

    Returns:
        int: Network size.
    """
    return int(cidr.split("/")[1])


def is_ip_in_cidr(ipaddr: str, cidr: str) -> bool:
    """Check if ip addr is in cidr range.

    Args:
        ipaddr (str): IP address to check.
        cidr (str): CIDR range to check.

    Returns:
        bool: True if ipaddr is in cidr range.
    """
    try:
        ip = ipaddress.ip_address(ipaddr)
        return ip in ipaddress.ip_network(cidr)
    except ValueError:
        return False


@cache.ttl_cache(maxsize=128, ttl=3600)
def get_tor_node_ip_addrs() -> List[str]:
    """Get a list of tor node ip addresses.

    Returns:
        List[str]: List of tor node ip addresses.
    """
    r = requests.get(TOR_PROJECT_NODE_ENDPOINT, timeout=DEFAULT_REQUEST_TIMEOUT)
    r.raise_for_status()
    return [line for line in r.text.splitlines() if line != ""]


def is_ip_tor_node(ipaddr: str, cached: bool = True) -> bool:
    """Check if ip addr is a tor node.

    Args:
        ipaddr (str): IP address to check.
        cached (bool): Force refresh of cache. Defaults to False.

    Returns:
        bool: True if ipaddr is a tor node.
    """
    if cached is False:
        get_tor_node_ip_addrs.clear_cache()

    nodes = get_tor_node_ip_addrs()
    return ipaddr in nodes


def get_ip_info(ipaddr: str) -> Optional[dict]:
    """Get ip address info.

    Args:
        ipaddr (str): IP address to get info for.

    Returns:
        Optional[dict]: IP address info.
    """
    if is_valid_ip_addr(ipaddr):
        r = requests.get(
            IPINFO_API_ENDPOINT % ipaddr, timeout=DEFAULT_REQUEST_TIMEOUT
        )
        r.raise_for_status()
        return r.json()

    return None


@cache.ttl_cache(maxsize=128, ttl=3600)
def get_aws_ip_ranges() -> Optional[Dict[str, Any]]:
    """Get the public ip ranges for AWS services."""
    try:
        r = requests.get(
            AWS_IP_RANGES_ENDPOINT, timeout=DEFAULT_REQUEST_TIMEOUT
        )
        r.raise_for_status()
        result = r.json()
        cidrs = []
        for key in ["ip_prefix", "ipv6_prefix"]:
            prefixes = result.get("%ses" % key.replace("ip_", ""))
            for prefix in prefixes:
                cidr = prefix.get(key)
                del prefix[key]
                prefix["prefix"] = cidr
            cidrs.extend(prefixes)
        return cidrs
    except requests.exceptions.HTTPError:
        return None


def is_aws_ip_addr(ipaddr: str) -> bool:
    """Check if ip addr is an AWS ip.

    Args:
        ipaddr (str): IP address to check.

    Returns:
        bool: True if ipaddr is an AWS ip.
    """
    try:
        ip = ipaddress.ip_address(ipaddr)
        return any(  # pragma: no cover
            ip in ipaddress.ip_network(cidr) for cidr in get_aws_ip_ranges()
        )
    except ValueError:
        return False


def get_cloudflare_range(endpoint: str) -> Optional[List[str]]:
    """Make request to endpoint and return only ip ranges."""
    results = []
    try:
        r = requests.get(endpoint, timeout=DEFAULT_REQUEST_TIMEOUT)
        r.raise_for_status()
        subnets = [line for line in r.text.split("\n") if line.strip() != ""]
        results.extend(subnets)
    except requests.exceptions.HTTPError:
        pass

    return results


@cache.ttl_cache(maxsize=128, ttl=3600)
def get_cloudflare_ip_ranges() -> Optional[Dict[str, Any]]:
    """Get the public ip ranges for Cloudflare services."""
    ip_ranges = []

    ip_ranges.extend(
        [
            *get_cloudflare_range(CLOUDFLARE_IPV4_RANGES_ENDPOINT),
            *get_cloudflare_range(CLOUDFLARE_IPV6_RANGES_ENDPOINT),
        ]
    )

    return ip_ranges


def is_cloudflare_ip_addr(ipaddr: str) -> bool:
    """Check if ip addr is a cloudflare ip.

    Args:
        ipaddr (str): IP address to check.

    Returns:
        bool: True if ipaddr is a fastly ip.
    """
    try:
        ip = ipaddress.ip_address(ipaddr)
        return any(  # pragma: no cover
            ip in ipaddress.ip_network(cidr)
            for cidr in get_cloudflare_ip_ranges()
        )
    except ValueError:
        return False


@cache.ttl_cache(maxsize=128, ttl=3600)
def get_fastly_ip_ranges() -> Optional[Dict[str, Any]]:
    """Get the public ip ranges for Fastly services."""
    try:
        r = requests.get(
            FASTLY_IP_RANGES_ENDPOINT, timeout=DEFAULT_REQUEST_TIMEOUT
        )
        r.raise_for_status()
        result = r.json()
        return [subnet for subnet in result.values()]
    except requests.exceptions.HTTPError:
        return []


def is_fastly_ip_addr(ipaddr: str) -> bool:
    """Check if ip addr is a fastly ip.

    Args:
        ipaddr (str): IP address to check.

    Returns:
        bool: True if ipaddr is a fastly ip.
    """
    try:
        ip = ipaddress.ip_address(ipaddr)
        return any(  # pragma: no cover
            ip in ipaddress.ip_network(cidr) for cidr in get_fastly_ip_ranges()
        )
    except ValueError:
        return False



"""Valkyrie Tools."""

__all__ = ["__version__", "DEFAULT_CONFIG_FILE", "config"]

__appname__ = "valkyrie_tools"
__version__ = "0.1.0"  # noqa

# from .configs import Config

# DEFAULT_CONFIG_FILE = os.path.join(os.path.expanduser("~"), ".valkyrie")
# config = Config(DEFAULT_CONFIG_FILE)


# CONFIG_DIR = user_config_dir(__appname__)
# CONFIG_FILE = os.path.join(CONFIG_DIR, f"{__appname__}")


from .config import Config


configs = Config(
    __appname__,
    defaults={
        "GLOBAL": {"virusTotalApiKey": ""},
    },
)



"""Package commons for valkyrie-tools."""

import os
import re
import sys
from functools import wraps
from typing import List, Optional, Tuple, Union

import click
from art import text2art

from .constants import (
    DOMAIN_REGEX,
    EMAIL_ADDR_REGEX,
    INTERACTIVE_MODE_PROMPT,
    IPV4_REGEX,
    IPV6_REGEX,
    URL_REGEX,
)
from .files import is_binary_file


__all__ = [
    "common_options",
    "print_version",
    "parse_input_methods",
    "extract_domains",
    "extract_ipv4_addrs",
    "extract_ipv6_addrs",
    "extract_ip_addrs",
    "extract_emails",
    "extract_urls",
]


def common_options(
    cmd_type: Union[click.Command, click.Group] = click.Command,
    name: str = "",
    description: str = "",
    version: str = "",
) -> Union[click.Command, click.Group]:
    """Decorator to add common options to command-line tools.

    Allows for built-in command definitions and the flags:
    -h, --help: Show help message and exit.
    -V, --version: Show version and exit.
    -vvv, --verbose: Enable verbose mode.
    -i, --interactive: Enable interactive mode.

    Args:
        cmd_type (Union[click.Command, click.Group]): Click command type.
        name (str): Name of the command.
        description (str): Description of the command.
        version (str): Version of the command in semantic versioning scheme.

    Returns:
        Union[click.Command, click.Group]: Click command function.
    """

    def decorator(
        func: Union[click.Command, click.Group]
    ) -> Union[click.Command, click.Group]:
        """Bind options to the command function."""

        @cmd_type(
            name=name,
            help=description,
            context_settings=dict(help_option_names=["-h", "--help"]),
            hidden=True,
        )
        # @click.option(
        #     "-i",
        #     "--input",
        #     "input",
        #     type=click.File('rb'),
        #     default=click.get_binary_stream('stdin'),
        #     help="Input file.",
        # )
        # @click.option(
        #     "-o",
        #     "--output",
        #     "output",
        #     type=click.File("w", encoding="utf8"),
        #     help="Output file.",
        #     # Doesn't work, the click stdout string does not work the same
        #     # as what's used in the click.s?echo functions.
        #     # default=click.get_text_stream("stdout"),
        #     default=None,
        # )
        @click.option(
            "-I",
            "--interactive",
            is_flag=True,
            help="Interactive mode.",
            default=False,
        )
        # @click.version_option(
        #    version,
        #    "-V",
        #    "--version",
        #    message="%(prog)s %(version)s",
        #    help="Show version and exit.",
        # )
        @click.argument(
            "values", nargs=-1, type=click.UNPROCESSED, required=False
        )
        @wraps(func)
        def wrapper(*args: List[str], **kwargs: List[str]) -> None:
            """Ensure decorated function's metadata is preserved."""
            return func(*args, **kwargs)

        return wrapper

    return decorator


def print_version(version: str) -> click.core.Command:
    """Print version and exit.

    Args:
        version (str): Version of the command in semantic versioning scheme.

    Returns:
        Callable: Click callback function.
    """

    def echo_version(
        ctx: click.Context,
        param: Union[click.Option, click.Parameter],  # noqa: B008
        value: str,
    ) -> None:
        """Print version and exit.

        Args:
            ctx (click.Context): Click context.
            param (Union[click.Option, click.Parameter]): Click option or
                parameter.
            value (str): Value of the option or parameter.
        """
        if not value or ctx.resilient_parsing:
            return

        click.echo(version)
        ctx.exit()

    return echo_version


def handle_file_input(
    file_path: str,
) -> Optional[str]:
    """Handle file input."""
    value = None

    # Safely avoid binary files.
    if is_binary_file(file_path) is True:
        raise OSError("Binary file provided, not a text file.")
    # Avoid directories.
    elif os.path.isdir(file_path) is True:
        raise OSError("Directory provided, not a file.")

    # Handle as normal file.
    elif os.path.isfile(file_path) is True:
        # Read the file, if possible.
        with open(file_path, encoding="utf-8") as file:
            value = file.read().strip()

    return value


def parse_input_methods(
    values: Tuple[str, ...],
    interactive: bool,
    ctx: click.Context,
) -> Tuple[str, ...]:
    """Parse input methods.

    This function will attempt to read from stdin if
    there's no values provided. If there's a "-" in the
    values, it will also read from stdin. If there's
    values provided, it will attempt to read from the
    file path provided in the values. If there's no file
    at the path provided, it will just use the value
    provided.

    Args:
        values (Tuple[str, ...]): Positional arguments.
        interactive (bool): Whether to enable interactive mode.
        ctx (click.Context): Click context.

    Returns:
        Tuple[str, ...]: Tuple of arguments.
    """
    args = ()

    # When interactive is enabled, drop to an input
    # stream.
    if interactive is True:
        # Convert the script name to ascii art as a header
        # and print it.
        click.echo(text2art(ctx.command.name))

        # Print the interactive prompt and read from stdin.
        click.echo(INTERACTIVE_MODE_PROMPT)
        stdin = sys.stdin.read().strip()

        # Only add the users input if it isn't empty.
        if stdin != "":
            values += (stdin,)

    # Check if data is being piped
    elif sys.stdin.isatty() is False:
        # Check if there's a "-" in the values
        if "-" in values:
            # Before reading from stdin, need to remove the
            # "-" from the values.
            values = tuple(v for v in values if v != "-")

        # Read data from pipe
        values += (sys.stdin.read().strip(),)

    # Iterate over each argument
    for arg in values:
        # Avoid empty strings since Path.exists return True
        # for them.
        if os.path.exists(arg) is True:
            arg = handle_file_input(arg)
            if arg is None:  # pragma: no cover
                continue

            args += (arg,)

        # Otherwise, tack on the arg to the tuple.
        else:
            args += (arg,)

    # Remove empty strings from args
    args = tuple(a for a in args if a != "")
    return args


# Regex extract functions
def _extract_regex(
    regex: re.Pattern, text: str, key: Optional[str], unique: bool = False
) -> List[str]:
    """Extract regex matches from text.

    Args:
        regex (re.Pattern): Regex pattern to use.
        text (str): Text to extract regex matches from.
        key (str, optional): Key to extract from regex match.
            Defaults to None.
        unique (bool): Whether to return unique matches only.
            Defaults to False.

    Returns:
        List[str]: List of matches.
    """
    matches = []
    for text_match in regex.finditer(text):
        match = text_match.group(key) if key else text_match.group()
        if unique and match in matches:
            continue

        matches.append(match)

    return matches


def extract_domains(text: str, unique=False) -> List[str]:
    """Extract domains from text.

    Args:
        text (str): Text to extract domains from.
        unique (bool, optional): Whether to return unique domains only.
            Defaults to False.

    Returns:
        List[str]: List of domains.
    """
    return _extract_regex(DOMAIN_REGEX, text, "domain", unique)


def extract_ipv4_addrs(text: str, unique=False) -> List[str]:
    """Extract IPv4 addresses from text.

    Args:
        text (str): Text to extract IPv4 addresses from.
        unique (bool, optional): Whether to return unique IPv4 addresses only.
            Defaults to False.

    Returns:
        List[str]: List of IPv4 addresses.
    """
    return _extract_regex(IPV4_REGEX, text, "ipv4", unique)


def extract_ipv6_addrs(text: str, unique=False) -> List[str]:
    """Extract IPv6 addresses from text.

    Args:
        text (str): Text to extract IPv6 addresses from.
        unique (bool, optional): Whether to return unique IPv6 addresses only.
            Defaults to False.

    Returns:
        List[str]: List of IPv6 addresses.
    """
    return _extract_regex(IPV6_REGEX, text, "ipv6", unique)


def extract_ip_addrs(text: str, unique=False) -> List[str]:
    """Extract IPv4 and IPv6 addresses from text.

    Args:
        text (str): Text to extract IPv4 and IPv6 addresses from.
        unique (bool, optional): Whether to return unique IPv4 and IPv6
            addresses only. Defaults to False.

    Returns:
        List[str]: List of IPv4 and IPv6 addresses.
    """
    return extract_ipv4_addrs(text, unique) + extract_ipv6_addrs(text, unique)


def extract_emails(text: str, unique=False) -> List[str]:
    """Extract email addresses from text.

    Args:
        text (str): Text to extract email addresses from.
        unique (bool, optional): Whether to return unique email addresses only.
            Defaults to False.

    Returns:
        List[str]: List of email addresses.
    """
    return _extract_regex(EMAIL_ADDR_REGEX, text, "email", unique)


def extract_urls(text: str, unique=False) -> List[str]:
    """Extract URLs from text.

    Args:
        text (str): Text to extract URLs from.
        unique (bool, optional): Whether to return unique URLs only. Defaults to False.

    Returns:
        List[str]: List of URLs.
    """
    return _extract_regex(URL_REGEX, text, "url", unique)



"""Whobe command-line script."""

import sys
from typing import Any, Dict, Optional, Tuple

import click

from .commons import (
    common_options,
    extract_domains,
    extract_ip_addrs,
    parse_input_methods,
)
from .constants import HELP_SHORT_TEXT, NO_ARGS_TEXT
from .ipaddr import get_net_size, is_valid_ip_addr
from .whois import get_ip_whois, get_whois


NO_WHOIS_MSG = "No whois data"


def print_ip_whois(whois: Optional[Dict[str, Any]]) -> None:  # pragma: no cover
    """Print results to the terminal."""
    if isinstance(whois, dict) is False:
        click.echo("No whois data.", err=True)
    else:
        click.echo(
            f'   ASN: {whois.get("asn", "Unknown")}'
            f' ({whois.get("asn_country_code", "Unknown")})'
        )
        click.echo("   CIDR: %s" % whois.get("asn_cidr", "Unknown"))
        click.echo("   Description: %s" % whois.get("asn_cidr", "Unknown"))
        click.echo("   Networks:")

        # Filter out networks that don't have a name or handle
        nets = filter(
            lambda x: isinstance(x, dict)
            and x.get("name") is not None
            and x.get("handle") is not None,
            whois.get("nets", []) or [],
        )
        for net in nets:
            # Print the network name and handle
            click.echo(f'      - {net.get("name", "Unknown")} ', nl=False)
            click.echo(f'({net.get("handle", "Unknown")})')

            # Print the netrange and total number of hosts
            click.echo(f'        CIDR: {net["cidr"]}')
            click.echo(f'        Netrange: ({net["range"]})', nl=False)
            click.echo(f' - {get_net_size(net["cidr"])} Hosts')

            # Print the address and location
            click.echo("        Address: ", nl=False)
            address = net.get("address", "Unknown")
            if isinstance(address, str) is True:
                address = address.replace("\n", " ").strip()

            click.echo(address, nl=False)
            city = net.get("city", "Unknown")
            if isinstance(city, str) is True:
                click.echo(f", {city}", nl=False)

            country = net.get("country", "Unknown")
            if isinstance(country, str) is True:
                click.echo(f", {country}", nl=False)

            postal_code = net.get("postal_code", "")
            if isinstance(postal_code, str) is True:
                click.echo(f" {postal_code}", nl=False)

            click.echo("")

            # Print the emails
            click.echo("        Emails: ", nl=False)
            if isinstance(net.get("emails", []), (list, tuple)) is True:
                click.echo("")
                for email in net["emails"]:
                    click.echo(f"          - {email}")
            else:
                click.echo(net.get("emails", "Unknown"))


def print_whois(whois: Optional[Dict[str, Any]]) -> None:  # pragma: no cover
    """Print results to the terminal."""
    if isinstance(whois, dict) is False:
        click.echo(NO_WHOIS_MSG, err=True)
    else:
        click.echo(
            f'   Registrar: {whois.get("registrar", "Unknown")} '
            f'({whois.get("org", "Unknown")})'
        )

        # status = whois.get("status", "Unknown")
        # if isinstance(status, (list, tuple)):
        #     status = ", ".join(
        #         list(set([text.split(" ")[0] for text in status]))
        #     )
        #
        # click.echo(f"   Status: {status}")

        click.echo("   Emails: ")
        emails = whois.get("emails", ["None"])
        if isinstance(emails, str):
            emails = [emails]

        for email in emails:
            click.echo(f"      - {email}")

        click.echo(f"   Name: {whois.get('name', 'Unknown')}")

        click.echo("   Address: ", nl=False)
        address = whois.get("address", "Unknown")
        city = whois.get("city", "Unknown")
        country = whois.get("country", "Unknown")
        state = whois.get("state", "Unknown")
        postal_code = whois.get("registrant_postal_code", "")
        click.echo(f"{address}, {city}, {state} {postal_code}, {country}")

        creation_date = whois.get("creation_date")
        click.echo(f"   Creation Date: {creation_date}")

        expiration_date = whois.get("expiration_date")
        click.echo(f"   Expiration Date: {expiration_date}")

        updated_date = whois.get("updated_date")
        click.echo(f"   Updated Date: {updated_date}")

        click.echo("   Name Servers: ")
        name_servers = whois.get("name_servers", ["None"])
        if isinstance(name_servers, (list, tuple)):
            for ns in whois["name_servers"]:
                click.echo(f"      - {ns}")
        else:
            click.echo(f"     - {name_servers}")


@common_options(
    cmd_type=click.command,
    name="whobe",
    description="Check whois on domains and ip addresses.",
    version="0.1.0",
)
@click.pass_context
def cli(
    ctx: click.Context,
    values: Tuple[str, ...],
    interactive: bool,
) -> None:
    """Check whois on domains and ip addresses."""
    values = parse_input_methods(values, interactive, ctx)
    if len(values) == 0:
        click.echo("Error: %s" % NO_ARGS_TEXT, err=True)
        click.echo(HELP_SHORT_TEXT.format(name=ctx.command.name), err=True)
        sys.exit(1)

    ip_addrs = extract_ip_addrs("\n".join(values), unique=True)
    domains = extract_domains("\n".join(values), unique=True)
    args = ip_addrs + domains

    for a in range(len(args)):
        arg = args[a]
        click.echo(f"> {arg}".format(arg))

        if is_valid_ip_addr(arg):
            whois = get_ip_whois(arg)
            print_ip_whois(whois)
        else:
            whois = get_whois(arg)
            print_whois(whois)

        # Print trailing newline
        if a < len(args) - 1:
            click.echo()


if __name__ == "__main__":  # pragma: no cover
    cli()



"""Module for DNS lookups."""

import re
from typing import List, Tuple

import dns.inet
import dns.name
import dns.query
import dns.rdatatype
import dns.resolver
import dns.tsig
import dns.tsigkeyring
import dns.update

from .ipaddr import is_valid_ip_addr


Timeout = dns.resolver.Timeout
DEFAULT_NAMESERVERS = [
    "1.1.1.1",
    "1.0.0.1",  # Cloudflare
    "8.8.8.8",
    "8.8.4.4",  # Google
    "9.9.9.9",
    "149.112.112.112",  # Quad9
    "208.67.222.222",
    "208.67.220.220",  # OpenDNS
    "8.26.56.26",
    "8.20.247.20",  # Comodo Secure
    "185.225.168.168",
    "185.228.169.168",  # CleanBrowsing
    "76.76.19.19",
    "76.223.122.150",  # Alternate
    "176.103.130.130",
    "176.103.130.131",  # AdGuard
    "64.6.64.6",
    "64.6.65.6",  # Verisign
]
DEFAULT_RECORD_TYPES = ["A", "AAAA", "MX", "CNAME", "PTR"]
RECORD_TYPES = [q.name for q in dns.rdatatype.RdataType]


def is_valid_record_type(record_type: str) -> bool:
    """Check if a given record type is valid.

    Args:
        record_type (str): Record type to check.

    Returns:
        bool: True if record_type is a valid record type.
    """
    try:
        return dns.rdatatype.from_text(record_type) is not None

    except (dns.rdatatype.UnknownRdatatype, AttributeError):
        return False


def get_rdns_record(ipaddr: str) -> list:
    """Get reverse DNS record for an IP address.

    Args:
        ipaddr (str): IP address to get reverse DNS record for.

    Returns:
        list: List of reverse DNS records.

    Raises:
        ValueError: If the IP address is invalid.
    """
    if is_valid_ip_addr(ipaddr) is False:
        raise ValueError(f"Invalid IP address: {ipaddr}")

    # Configure DNS resolver with default name servers
    resolver = dns.resolver.Resolver(configure=False)
    resolver.nameservers = DEFAULT_NAMESERVERS
    resolver.search = []

    try:
        # Retrieve reverse DNS records for the IP address
        records = resolver.resolve(dns.reversename.from_address(ipaddr), "PTR")
        return [(record.rdtype.name, record.to_text()) for record in records]
    except (
        dns.exception.SyntaxError,
        dns.resolver.NXDOMAIN,
        dns.resolver.NoAnswer,
    ):
        return []


def get_dns_record(
    domain: str,
    record_type: str,
    nameservers: List[str] = DEFAULT_NAMESERVERS,
) -> List[Tuple[str, str]]:
    """Get DNS record for a domain.

    Args:
        domain (str): Domain to get DNS record for.
        record_type (str): Record type to get.
        nameservers (list[str], optional): List of name servers to use.

    Raises:
        ValueError: If the record type is invalid.

    Returns:
        list[Tuple[str, str]]: List of DNS records.
    """
    record_type = record_type.upper()
    if is_valid_record_type(record_type) is False:
        raise ValueError(f"Invalid record type: {record_type}")

    # Configure DNS resolver with default name servers
    resolver = dns.resolver.Resolver(configure=False)
    resolver.nameservers = nameservers
    resolver.search = []

    resolves = []

    if record_type == "PTR" and is_valid_ip_addr(domain) is True:
        resolves.extend(get_rdns_record(domain))
    else:
        try:
            # Retrieve DNS records for the domain
            records = resolver.resolve(domain, record_type)  # Example: A record
            for record in records:
                type_name = record.rdtype.name
                value = record.to_text()

                if type_name == "MX":
                    value = re.sub(r"(\s+?)?\d+(\s+?)?", "", value)
                    if value == ".":
                        continue

                resolves.append((type_name, value))
        except (
            dns.exception.SyntaxError,
            dns.resolver.NXDOMAIN,
            dns.resolver.NoAnswer,
        ):
            pass

    return resolves


def get_dns_records(
    domain: str, record_types: List[str] = DEFAULT_RECORD_TYPES
) -> List[List[Tuple[str, str]]]:
    """Get DNS records for a domain.

    Args:
        domain (str): Domain to get DNS records for.
        record_types (List[str], optional): List of record types to get.
            Defaults to DEFAULT_RECORD_TYPES.

    Returns:
        List[List[Tuple[str, str]]]: List of DNS records.
    """
    records = []
    for record_type in record_types:
        records.extend(get_dns_record(domain, record_type))

    return records



"""Whois utility functions."""

from time import sleep
from typing import Optional, Union

import ipwhois
import whois
from ipwhois import IPWhois


__all__ = [
    "get_whois",
    "get_ip_whois",
]

WHOIS_MAX_RETRIES = 3


def get_whois(
    domain: str,
) -> Optional[Union[whois.parser.WhoisCom, whois.parser.WhoisEntry]]:
    """Get whois information for a domain."""
    w = None
    try:
        attempts = 0
        while attempts < WHOIS_MAX_RETRIES:
            attempts += 1
            w = whois.whois(domain)

            if w is not None and domain in w.get("domain", []):
                break

            sleep(0.25 * attempts)
    except whois.parser.PywhoisError:
        w = None

    return w


def get_ip_whois(ipaddr: str) -> Optional[dict]:
    """Get ip whois information for a valid ip addr."""
    results = None
    try:
        ipw = IPWhois(ipaddr)
        results = ipw.lookup_whois(
            retry_count=0, asn_methods=["dns", "whois", "http"]
        )
        # results = ipw.lookup_rdap()
    except (ValueError, ipwhois.exceptions.IPDefinedError):
        results = None

    return results



"""Cache module for persistent data."""

from functools import lru_cache, update_wrapper
from math import floor
from time import time
from typing import Any, Callable


def _ttl_hash_gen(seconds: int) -> int:
    """Generates a hash value based on the elapsed time.

    Since the start of the generator, divided by the specified number of seconds.

    Args:
        seconds (int): The number of seconds to divide the elapsed time by.

    Yields:
        int: The hash value generated based on the elapsed time.
    """
    start_time = time()
    while True:
        yield floor((time() - start_time) / seconds)


class Cache:
    """A decorator class for caching function results."""

    @staticmethod
    def memoize(fn: Callable) -> Callable:
        """Decorator for caching function results.

        Memoize decorator that caches the return value of a function
        based on its arguments.

        Args:
            fn (Callable): The function to be memoized.

        Returns:
            The memoized function.
        """
        memo = {}

        def wrapper(*args: Any) -> Any:
            """A wrapper that caches the result of the function."""
            if args in memo:
                return memo[args]
            else:
                rv = fn(*args)
                memo[args] = rv
                return rv

        return wrapper

    @staticmethod
    def ttl_cache(
        maxsize: int = 128, typed: bool = False, ttl: int = -1
    ) -> Callable:
        """Decorator that adds TTL caching functionality to a function.

        Args:
            maxsize (int): The maximum number of function calls to
                cache. Defaults to 128.
            typed (bool): Whether to differentiate between
                arguments of different types. Defaults to False.
            ttl (int): The time-to-live (in seconds) for the cached
                function results. Defaults to -1, which means no expiration.

        Returns:
            Callable: The decorated function.
        """
        # Any ttl that's 0 or less, set to max.
        if ttl <= 0:
            ttl = 65536  # pragma: no cover

        hash_gen = _ttl_hash_gen(ttl)

        def wrapper(func: Callable) -> Callable:
            """A decorator that adds caching functionality to a function.

            Args:
                func (Callable): The function to be decorated.

            Returns:
                Callable: The decorated function.
            """

            @lru_cache(maxsize, typed)
            def ttl_func(ttl_hash: str, *args, **kwargs) -> Any:
                """Caches the result of 'func' function based on the arguments.

                Args:
                    ttl_hash (str): Hash used to identify the cached result.
                    *args: Positional arguments passed to the 'func' function.
                    **kwargs: Keyword arguments passed to the 'func' function.

                Returns:
                    The result of the 'func' function.
                """
                return func(*args, **kwargs)

            def wrapped(*args, **kwargs) -> Any:
                """A wrapper that generates a hash value and calls the ttl_func.

                This function is a wrapper that generates a hash value and
                calls the ttl_func with the generated hash value, along with
                the provided arguments and keyword arguments.

                Args:
                    *args: Variable length argument list.
                    **kwargs: Arbitrary keyword arguments.

                Returns:
                    Any: The result returned by ttl_func.
                """
                th = next(hash_gen)
                return ttl_func(th, *args, **kwargs)

            def clear_cache() -> Any:
                """Clears the cache used by the ttl_func."""
                ttl_func.cache_clear()

            wrapped.clear_cache = clear_cache

            return update_wrapper(wrapped, func)

        return wrapper


cache = Cache()



"""Command-line interface."""

from typing import Optional

import click

from . import __version__, configs


# @common_options(
#    cmd_type=click.group,
#    name="valkyrie",
#    description="Vallyrie Tools Toolkit Interface.",
#    version=__version__,
# )
@click.group(
    help="Valkyrie Toolkit Interface.",
    context_settings={"help_option_names": ["-h", "--help"]},
)
@click.version_option(version=__version__, prog_name="valkyrie")
def cli() -> None:
    """Valkyrie command."""
    pass  # pragma: no cover


@cli.group(name="config")
def config_group() -> None:
    """Configuration management."""
    pass  # pragma: no cover


@config_group.command(name="set")
@click.argument("key", metavar="key")
@click.argument("value", metavar="value")
def set_config(key: str, value: str) -> None:
    """Set a configuration value."""
    configs.set("GLOBAL", key, value)
    # click.echo(f"Config updated: {configs.config_file}")
    # click.echo(f"Config updated: {configs.config_file}")
    click.echo(f"Updated {key}.")


@config_group.command(name="get")
@click.argument("key", metavar="key")
def get_config(key: str) -> None:
    """Get a configuration value."""
    value = configs.get("GLOBAL", key)
    click.echo(f"{key}: {value}")


@config_group.command(name="delete")
@click.argument("key", metavar="key")
def delete_config(key: str) -> None:
    """Delete a configuration value."""
    configs.remove_option("GLOBAL", key)
    click.echo(f"Config updated: {configs.config_file}")
    click.echo(f"Deleted {key}.")


@config_group.command(name="list")
@click.argument("key", metavar="key", required=False)
def list_config(key: Optional[str] = None) -> None:
    """List configuration values."""
    if key is not None:
        for key, value in configs.config.items("GLOBAL"):
            if key.lower() == key.lower():  # pragma: no cover
                click.echo(f"{key}: {value}")
    else:
        for key, value in configs.config.items("GLOBAL"):
            click.echo(f"{key}: {value}")  # pragma: no cover


if __name__ == "__main__":
    cli()  # pragma: no cover



"""Config module."""

import configparser
import os
from typing import Any, Dict

from appdirs import user_config_dir


class Config:
    """A class for managing a configuration file in the user's config directory.

    Attributes:
        config_name (str): The name of the configuration file.
        config_file (str): The full path to the configuration file.
        config (configparser.ConfigParser): The configparser for the
            configuration file.
    """

    def __init__(
        self,
        config_name: str,
        defaults: Dict[str, Dict[str, Any]] = None,
    ):
        """Initialize the Config object.

        Args:
            config_name (str): The name of the configuration file.
            defaults (dict, optional): The configuration default values.
                Defaults to None.
        """
        self.config_name = config_name
        self.config_file = os.path.join(user_config_dir(), f".{config_name}")
        self.config = configparser.ConfigParser()
        self.defaults = defaults

        if self.read() == [] and self.defaults is not None:
            self.set_defaults()

    def read(self):
        """Load the configuration file."""
        return self.config.read(self.config_file)

    def save(self):
        """Save the configuration file."""
        with open(self.config_file, "w") as f:
            self.config.write(f)

    def get(self, section: str, option: str, fallback: Any = None) -> Any:
        """Get the value of an option in a section.

        Args:
            section (str): The section of the option.
            option (str): The name of the option.
            fallback (Any): The fallback value if the option is not found.
                Defaults to None.

        Returns:
            Any: The value of the option.
        """
        return self.config.get(section, option, fallback=fallback)

    def set(self, section: str, option: str, value: Any):
        """Set the value of an option in a section.

        Args:
            section (str): The section of the option.
            option (str): The name of the option.
            value (Any): The value of the option.
        """
        if not self.config.has_section(section):
            self.config.add_section(section)

        self.config.set(section, option, str(value))
        self.save()

    def remove_section(self, section: str):
        """Remove a section from the configuration file.

        Args:
            section (str): The section to remove.
        """
        self.config.remove_section(section)
        self.save()

    def remove_option(self, section: str, option: str):
        """Remove an option from a section.

        Args:
            section (str): The section of the option.
            option (str): The name of the option.
        """
        self.config.remove_option(section, option)
        self.save()

    def set_defaults(self):
        """Set default values for the configuration."""
        for section, options in self.defaults.items():
            for option, value in options.items():
                if not self.config.has_section(section):
                    self.config.add_section(section)
                self.config.set(section, option, str(value))
        self.save()



"""File utilities."""

import os
from typing import Optional, Union


__all__ = [
    "path_exists",
    "is_binary_file",
    "is_file_descriptor",
    "read_file",
]


def path_exists(path: str) -> bool:
    """Check if the file is a regular file.

    Args:
        path (str): File path.

    Returns:
        bool: True if path is a regular file.
    """
    try:
        return os.path.exists(path)
    except Exception:  # pragma: no cover
        return False


def is_binary_file(path: str) -> bool:
    """Check if the file is a binary file.

    Args:
        path (str): File path.

    Returns:
        bool: True if path is a binary file.
    """
    try:
        with open(path, "rb") as f:
            # Read the first 1024 bytes
            data = f.read(1024)

        return any(byte < 9 or byte in (11, 12, 14, 15, 127) for byte in data)
    except OSError:  # pragma: no cover
        return False


def is_file_descriptor(path: Union[str, int]) -> bool:
    """Check if arg is a file descriptor.

    Args:
        path (Union[str, int]): File path or file descriptor.

    Returns:
        bool: True if path is a file descriptor.
    """
    if isinstance(path, int):
        try:
            os.fstat(path)
            return True
        except Exception:
            return False
    elif isinstance(path, str):
        try:
            # On Linux, /dev/fd/x is a symlink to /proc/self/fd/x and
            # /proc/self/fd/x is a pseudo-symlink to the file open on
            # fd x.
            if (
                os.path.exists(path) is True
                # and os.path.islink(path) is True
                and "/fd/" in path
            ):
                return True
        except Exception:
            return False
    else:  # pragma: no cover
        return False


def read_file(path: str) -> Optional[str]:
    """Read file contents.

    Args:
        path (str): Path to file.

    Raises:
        OSError: If path does not exist.
        OSError: If path is a directory.
        OSError: If path is not a file.

    Returns:
        Optional[str]: File contents.
    """
    if os.path.exists(path) is False:
        raise OSError("Path does not exist. %s" % path)

    # Resolve symlinks
    if os.path.islink(path) is True:
        # Avoid trying to resolve file descriptors
        # as symlinks.
        if "/fd/" in path:  # pragma: no cover
            fd_sp = path.split("/")
            fd_num = int(fd_sp[-1])
            with os.fdopen(fd_num, mode="r", encoding="utf-8") as fd:
                return fd.read()

        # Resolve symlinks to non-descriptors
        path = os.readlink(path)  # pragma: no cover

    # Read simple file
    if os.path.isfile(path) is True:
        with open(path, encoding="utf-8") as f:
            return f.read()

    # Raise error if path is a directory
    elif os.path.isdir(path) is True:
        raise OSError("Path is a directory. %s" % path)

    # Raise error if path is not a file
    else:
        raise OSError("Path is not a file. %s" % path)



"""Exception modules."""

REQUESTS_SSL_ERROR_MESSAGE = "SSL error."
REQUESTS_INVALID_URL_ERROR_MESSAGE = "Invalid URL."
REQUESTS_CONNECTION_ERROR_MESSAGES = {
    "Failed to resolve": "Failed to resolve host.",
    "Name or service not known": "Name or service not known.",
    "Temporary failure in name resolution": "Temporary failure in name resolution.",
    "nodename nor servname provided, or not known": (
        "Node name nor server name provided, or not known."
    ),
    "Connection refused": "Connection refused.",
    "Connection timed out": "Connection timed out.",
    "Connection reset by peer": "Connection reset by peer.",
    "Connection closed": "Connection closed.",
    "Connection aborted": "Connection aborted.",
    "Connection broken": "Connection broken.",
}
REQUESTS_UNHANDLED_CONNECTION_ERROR_MESSAGE = "Unhandled connection failure."
REQUESTS_TIMEOUT_ERROR_MESSAGE = "Failed to connect, timeout reached."
REQUESTS_TOO_MANY_REDIRECTS_ERROR_MESSAGE = "Too many redirects."



"""Command-line script for checking ip address info."""

import sys
from typing import Tuple

import click

from .commons import common_options, extract_ip_addrs, parse_input_methods
from .constants import HELP_SHORT_TEXT, NO_ARGS_TEXT
from .ipaddr import get_ip_info, is_private_ip


PRIVATE_IP_SKIP_MESSAGE = "Skipped, private ip address."


@common_options(
    cmd_type=click.command,
    name="ipcheck",
    description="Get ip address info.",
    version="0.1.0",
)
@click.pass_context
def cli(
    ctx: click.Context,
    values: Tuple[str, ...],
    interactive: bool,
) -> None:
    """Get ip address info."""
    values = parse_input_methods(values, interactive, ctx)
    if len(values) == 0:
        click.echo("Error: %s" % NO_ARGS_TEXT, err=True)
        click.echo(HELP_SHORT_TEXT.format(name=ctx.command.name), err=True)
        sys.exit(1)
    args = extract_ip_addrs("\n".join(values), unique=True)

    for a in range(len(args)):
        ipaddr = args[a]
        click.echo(f"> {ipaddr}".format(ipaddr))

        if is_private_ip(ipaddr):
            click.echo("  %s" % PRIVATE_IP_SKIP_MESSAGE)
            continue

        ipinfo = get_ip_info(ipaddr)

        key_width = max([len(k) for k in ipinfo.keys()]) + 1

        for k, v in ipinfo.items():
            click.echo(f"  {k:<{key_width}}: {v}")

        # Print trailing newline
        if a < len(args) - 1:
            click.echo()



"""Constants modules."""

import re


__all__ = [
    "HELP_SHORT_TEXT",
    "NO_ARGS_TEXT",
    "INVALID_ARG_TEXT",
    "INVALID_FLAG_TEXT",
    "BINARY_FILE_READ_ERROR",
    "YES_NO_PROMPT",
    "INTERACTIVE_MODE_PROMPT",
    "EMPTY_ARGS_NOT_ALLOWED",
    "DEFAULT_CATEGORIZED_HEADERS",
    "DOMAIN_REGEX_TEXT",
    "DOMAIN_REGEX",
    "IPV6_REGEX_TEXT",
    "IPV6_REGEX",
    "IPV4_REGEX_TEXT",
    "IPV4_REGEX",
    "EMAIL_ADDR_REGEX_TEXT",
    "EMAIL_ADDR_REGEX",
    "URL_REGEX_TEXT",
    "URL_REGEX",
]

# Errors
HELP_SHORT_TEXT = "Try '{name} -h' or '{name} --help' for more information."
HELP_VERSION_TEXT = "Show the version and exit."
NO_ARGS_TEXT = "No arguments provided."
INVALID_ARG_TEXT = "Invalid %s"
INVALID_FLAG_TEXT = "Invalid flag %s"
BINARY_FILE_READ_ERROR = "Cannot read binary files"

# Prompts
YES_NO_PROMPT = "%s [y/N]: "
INTERACTIVE_MODE_PROMPT = (
    "Enter/paste your text and Ctrl-D (Cmd-D on Mac or Ctrl-Z on Windows) to "
    "save it."
)
EMPTY_ARGS_NOT_ALLOWED = "Input args cannot be empty."

# Http default headers
DEFAULT_CATEGORIZED_HEADERS = {
    "AWS": {
        "Cloudfront": [
            # https://docs.aws.amazon.com/AmazonCloudFront/latest/DeveloperGuide/adding-cloudfront-headers.html
            "X-Amz-Cf-Id",
            "X-Amz-Cf-Pop",
            "X-Amzn-Trace-Id",
            "X-Cache",
            "X-Cache-Hits",
            "X-Served-By",
            "X-Timer",
            "X-XSS-Protection",
        ],
        "API Gateway": [
            "x-amzn-RequestId",
            "x-amzn-ErrorType",
            "x-amz-apigw-id",
            "X-Amz-Cf-Id",
            "X-Amzn-Trace-Id",
            "x-amzn-Remapped-Date",
        ],
        # AWS ELB
        "ELB": [
            # https://docs.aws.amazon.com/elasticloadbalancing/latest/userguide/how-elastic-load-balancing-works.html#request-routing
            # https://medium.com/@lancers/amazon-api-gateway-explaining-http-proxy-in-http-api-3ea0afe6b03c#:%7E:text=RequestId%20and%20API%20Gateway%20emitted%20custom%20headers
            "X-Forwarded-For",
            "X-Forwarded-Proto",
            "X-Forwarded-Port",
        ],
        "S3": [
            # https://docs.aws.amazon.com/AmazonS3/latest/API/RESTCommonResponseHeaders.html
            "Server",
            "x-amz-delete-marker",
            "x-amz-id-2",
            "x-amz-request-id",
            "x-amz-server-side-encryption",
            "x-amz-version-id",
        ],
    },
    "Fastly": {
        # https://developer.fastly.com/reference/http/http-headers
        "Fastly": [
            "Fastly-Cachetype",
            "Fastly-Client",
            "Fastly-Client-IP",
            "Fastly-Cookie-Overflow",
            "Fastly-Debug",
            "Fastly-Debug-Digest",
            "Fastly-Debug-Path",
            "Fastly-Debug-TTL",
            "Fastly-FF",
            "Fastly-Force-Shield",
            "Fastly-IO-Error",
            "Fastly-IO-Info",
            "Fastly-IO-Warning",
            "Fastly-Key",
            "Fastly-No-Shield",
            "Fastly-SSL",
            "Fastly-Stats",
            "X-Fastly-Imageopto-API",
            "X-Fastly-Imageopto-Montage",
            "X-Fastly-Imageopto-Overlay",
        ],
    },
    "Cloudflare": {
        # https://developers.cloudflare.com/fundamentals/reference/http-request-headers/
        "Cloudflare": [
            "Accept-Encoding",
            "CF-Connecting-IP",
            "CF-Connecting-IP",
            "CF-Connecting-IPv6",
            "CF-EW-Via",
            "CF-Pseudo-IPv4",
            "True-Client-IP",
            "X-Forwarded-For",
            "X-Forwarded-Proto",
            "CF-RAY",
            "CF-IPCountry",
            "CF-Visitor",
            "CDN-Loop",
            "CF-Worker",
            "Connection",
        ],
    },
    "Akamai": {
        # https://developer.akamai.com/legacy/introduction/Client_Request_Headers.html
        "Akamai": [
            "Akamai-Origin-Hop",
            "X-Akamai-Request-ID",
            "X-Cache",
            "X-Cache-Remote",
            "X-Check-Cacheable",
            "X-True-Cache-Key",
            "X-Cache-Key",
            "X-Serial",
        ]
    },
}
DEFAULT_REQUEST_TIMEOUT = 15

# Regex Patterns
DOMAIN_REGEX_TEXT = (
    r"(?P<domain>(?:(?P<subdomain>[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?\.)+"
    r"(?P<root>[a-zA-Z]{2,63})))"
)
DOMAIN_REGEX = re.compile(DOMAIN_REGEX_TEXT, re.I | re.M)

IPV6_REGEX_TEXT = (
    r"(?P<ipv6>"
    r"(?:[0-9a-fA-F]{1,4}:){7}[0-9a-fA-F]{1,4}"  # Standard IPv6
    r"|::(?:[0-9a-fA-F]{1,4}:)*[0-9a-fA-F]{1,4}"  # Abbreviated IPv6 with "::"
    r"|(?:[0-9a-fA-F]{1,4}:)*[0-9a-fA-F]{1,4}::"  # Abbreviated IPv6 with "::"
    # Abbreviated IPv6 with "::"
    r"|(?:[0-9a-fA-F]{1,4}:){1,6}::(?:[0-9a-fA-F]{1,4}:)*[0-9a-fA-F]{1,4}"
    # Abbreviated IPv6 with "::"
    r"|(?:[0-9a-fA-F]{1,4}:)*[0-9a-fA-F]{1,4}::(?:[0-9a-fA-F]{1,4}:)*"
    r"[0-9a-fA-F]{1,4}"
    # Abbreviated IPv6 ending with ":"
    r"|(?:[0-9a-fA-F]{1,4}:){1,5}:[0-9a-fA-F]{1,4}"
    r"|:[0-9a-fA-F]{1,4}:"  # IPv6 with a double colon in the middle
    r")"
)
IPV6_REGEX = re.compile(IPV6_REGEX_TEXT, re.I | re.M)

IPV4_REGEX_TEXT = (
    r"(?P<ipv4>"
    r"(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)(?:\.(?:25[0-5]|2[0-4][0-9]|"
    r"[01]?[0-9][0-9]?)){3}"
    r")"
)
IPV4_REGEX = re.compile(IPV4_REGEX_TEXT, re.I | re.M)

EMAIL_ADDR_REGEX_TEXT = (
    r"(?P<email>(?P<alias>[a-zA-Z0-9._-]+)\+?(?P<user>[a-zA-Z0-9._-]+)@"
    + DOMAIN_REGEX.pattern
    + r")"
)
EMAIL_ADDR_REGEX = re.compile(EMAIL_ADDR_REGEX_TEXT, re.I | re.M)

URL_REGEX_TEXT = (
    r"(?P<url>"
    r"(?P<protocol>[a-zA-Z0-9._-]+):\/\/"  # Protocol
    # Username and Password
    r"(?:(?P<username>[^:@]+)(?::(?P<password>[^@]+))?@)?"
    r"(?:"
    + DOMAIN_REGEX_TEXT
    + r"|"
    + IPV4_REGEX_TEXT
    + r"|"
    + IPV6_REGEX_TEXT  # Domain, IPv4, or IPv6
    + r")"
    r"(?::(?P<port>\d+))?"  # Port
    r"(?P<uri>[^?#]*)?"  # URI (anything before the query or fragment)
    r")"
)
URL_REGEX = re.compile(URL_REGEX_TEXT, re.I | re.M)



"""DNS lookup command-line script."""

import sys
from typing import List, Tuple

import click

from .commons import (
    common_options,
    extract_domains,
    extract_ip_addrs,
    parse_input_methods,
)
from .constants import HELP_SHORT_TEXT, NO_ARGS_TEXT
from .dns import DEFAULT_RECORD_TYPES, RECORD_TYPES, get_dns_records


def print_results(arg: str, results: dict) -> None:
    """Print the results of the DNS query."""
    click.echo(f"> {arg}".format(arg))
    rtype_len = max([len(rtype) for rtype, _ in results])
    for key, value in results:
        click.echo(f"  {key:{rtype_len}}: {value}")


@common_options(
    cmd_type=click.command,
    name="dnscheck",
    description="Check DNS records for domains and IP addresses.",
    version="0.1.0",
)
@click.option(
    "-t",
    "--rtypes",
    "record_types",
    help="DNS record type to query.",
    type=click.Choice(RECORD_TYPES),
    multiple=True,
    default=DEFAULT_RECORD_TYPES,
)
@click.pass_context
def cli(
    ctx: click.Context,
    values: Tuple[str, ...],
    interactive: bool,
    record_types: List[str],
) -> None:  # noqa: C901
    """Check whois on domains and ip addresses."""
    args = parse_input_methods(values, interactive, ctx)

    if len(args) == 0:
        click.echo("Error: %s" % NO_ARGS_TEXT, err=True)
        click.echo(HELP_SHORT_TEXT.format(name=ctx.command.name), err=True)
        sys.exit(1)

    ip_addrs = extract_ip_addrs("\n".join(values), unique=True)
    domains = extract_domains("\n".join(values), unique=True)
    args = list(set(ip_addrs + domains))

    for a in range(len(args)):
        arg = args[a]
        results = get_dns_records(arg, record_types=record_types)
        print_results(arg, results)

        # Print trailing newline
        if a < len(args) - 1:  # pragma: no cover
            click.echo()



"""Logger module."""

import logging


__all__ = [
    "logging",
    "log_message_format",
    "log_date_format",
    "log_levels",
    "get_verbosity",
    "set_log_level",
    "get_log_level",
    "setup_logger",
]

# log_message_format = "%(asctime)s %(levelname)s %(message)s"
log_message_format = "%(message)s"
log_date_format = "%Y-%m-%d %H:%M:%S"
log_levels = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]


def get_verbosity(verbose: int) -> int:
    """Log level getter function."""
    # Define log level based on verbosity count
    log_level = int(
        log_levels[
            min(verbose, len(log_levels) - 1)
        ]  # Max verbosity is DEBUG (3)
    )
    # Set log level
    return log_level


def set_log_level(logger: logging.Logger, log_level: int) -> None:
    """Log level setter function."""
    # Set log level
    logger.setLevel(log_level)


def get_log_level(logger: logging.Logger) -> int:
    """Log level getter function."""
    log_level = logger.getEffectiveLevel()
    # level_str = logging.getLevelName(log_level)
    return log_level


def setup_logger(verbose: int) -> logging.Logger:
    """Wrapper function with built-in verbosity handling."""
    logger = logging.getLogger(__name__)
    # Check if a StreamHandler already exists
    if not logger.handlers:
        handler = logging.StreamHandler()
        # get basename of script and name is filename
        # without extension
        formatter = logging.Formatter(log_message_format, log_date_format)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    # Set log level
    set_log_level(logger, verbose)
    return logger
